--[[
	main.lua
		Calculate and display item scores on tooltips.
--]]

ItemScore = LibStub("AceAddon-3.0"):NewAddon("ItemScore", 'AceConsole-3.0')

-- Libraries --
local TipHooker = LibStub("LibTipHooker-1.1")
local StatLogic = LibStub("LibStatLogic-1.2")

-- Local data --

--[[ ItemScore functions ]]--

function ItemScore:InitStats()
	-- store list of StatLogic stat names to provide scores for
	self.supportedStats = 
	{
		["EMPTY_SOCKET_RED"] = true,
		["EMPTY_SOCKET_YELLOW"] = true,
		["EMPTY_SOCKET_BLUE"] = true,
		["EMPTY_SOCKET_META"] = true,
		["EMPTY_SOCKET_PRISMATIC"] = true,

		["STR"] = true,
		["AGI"] = true,
		["STA"] = true,
		["INT"] = true,
		["SPI"] = true,
		["ARMOR"] = true,
		["ARMOR_BONUS"] = true,

		["FIRE_RES"] = true,
		["NATURE_RES"] = true,
		["FROST_RES"] = true,
		["SHADOW_RES"] = true,
		["ARCANE_RES"] = true,

		["BLOCK_VALUE"] = true,

		["AP"] = true,
		["RANGED_AP"] = true,
		["FERAL_AP"] = true,
		
		["HEAL"] = true,

		["SPELL_DMG"] = true,
		["SPELL_POWER"] = true,
		
		["SPELLPEN"] = true,

		["HEALTH"] = true,
		["MANA"] = true,
		["HEALTH_REGEN"] = true,
		["MANA_REGEN"] = true,

		["MAX_DAMAGE"] = true,
		["DPS"] = true,

		["DEFENSE_RATING"] = true,
		["DODGE_RATING"] = true,
		["PARRY_RATING"] = true,
		["BLOCK_RATING"] = true,
		["MELEE_HIT_RATING"] = true,
		--["RANGED_HIT_RATING"] = true,
		--["SPELL_HIT_RATING"] = true,
		["MELEE_CRIT_RATING"] = true,
		--["RANGED_CRIT_RATING"] = true,
		--["SPELL_CRIT_RATING"] = true,
		["RESILIENCE_RATING"] = true,
		["MELEE_HASTE_RATING"] = true,
		--["RANGED_HASTE_RATING"] = true,
		--["SPELL_HASTE_RATING"] = true,
		["EXPERTISE_RATING"] = true,
		["ARMOR_PENETRATION_RATING"] = true,
		["MASTERY_RATING"] = true,
		
		["VERSATILITY"] = true,
		["MULTISTRIKE"] = true,
		["AVOIDANCE"] = true,
		["LEECH"] = true,
		["RUN_SPEED"] = true,
	}
	self.reverseStats = {}
	
	-- now set the values of each index to readable values from stat logic
	for i in pairs(self.supportedStats) do
		local statName = StatLogic:GetStatNameFromID(i)
		if not statName then
			statName = i
		end
		self.supportedStats[i] = statName
		self.reverseStats[statName] = i
	end
	
	-- trinket value table
	-- REVISIT - there's probably a way to get this info from StatLogic
	-- but I can't figure it out
	self.trinketStats = {
		[string.lower(ITEM_MOD_HASTE_RATING_SHORT)] = "MELEE_HASTE_RATING",
		[string.lower(ITEM_MOD_CRIT_RATING_SHORT)] = "MELEE_CRIT_RATING",
		[string.lower(ITEM_MOD_MASTERY_RATING_SHORT)] = "MASTERY_RATING",
	}
end

function ItemScore:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("ItemScoreDB")
	self:InitStats()
	self:CheckDB()
	
	self.options = {}
	self:InitOptions()
	self:RegisterSlashCommands()
	
	LibStub("AceConfig-3.0"):RegisterOptionsTable("ItemScore", self.options)
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("ItemScore", "ItemScore")
end

function ItemScore:OnEnable()
    -- Called when the addon is enabled
    
    -- Hook item tooltips
	TipHooker:Hook(self.ProcessTooltip, "item")
end

function ItemScore:OnDisable()
    -- Called when the addon is disabled
    
    -- Unhook item tooltips
	TipHooker:Unhook(self.ProcessTooltip, "item")
end

--[[ Local Functions ]]--

--[[
	The Code
--]]

function ItemScore:GetTotalCurrentStats(currentStats, stat)
	-- calculate total stats of currently worn equipment
	if currentStats[stat] then return end
	
	currentStats[stat] = 0
	
	for i = 0, 19 do
		local link = GetInventoryItemLink("player", i)
		local itemTable = {}
		StatLogic:GetSum(link, itemTable)
		if itemTable[stat] then
			currentStats[stat] = currentStats[stat] + itemTable[stat]
		end
	end
end

local function getCappedValue(cap, currentTotal, value)
	-- returns the amount of value counted such that currentTotal + value <= cap
	local statLeft = math.max(0, cap - currentTotal)
	
	return math.min(value, statLeft)
end

function ItemScore:DoCategoriesMatch(setCategories, adjustmentCategories)
	-- All matches anything
	if not setCategories["All"] and not setCategories["Melee"] and 
	   	not setCategories["Ranged"] and 
	   	not setCategories["Healer"] then
	   	return false
	elseif not adjustmentCategories["All"] and 
		not adjustmentCategories["Melee"] and 
		not adjustmentCategories["Ranged"] and 
	   	not adjustmentCategories["Healer"] then
	   	return false
	elseif setCategories["All"] or adjustmentCategories["All"] then
	   return true
    end
    
    -- iterate through categories, returning true if any match
    for k in pairs(setCategories) do
    	if setCategories[k] and adjustmentCategories[k] then
    		return true
		end
	end
	
	return false
end

function ItemScore:ApplyAdjustments(itemTable, itemAdj, scoreSet)
	local applied = false
	if not itemTable or not itemAdj then
		return itemTable, applied
	end
	
	local setCategories = scoreSet.trinketCategories
	local ret = itemTable
	local tableCopied = false
	for _, v in pairs(itemAdj) do
		if self:DoCategoriesMatch(setCategories, v.categories) then
			-- apply adjustment
			-- need to copy item table
			if not tableCopied then
				ret = deepcopy(itemTable)
				tableCopied = true
			end
			--DEFAULT_CHAT_FRAME:AddMessage("Matched entry: stat - " .. (v.statToAdjust or "nil") .. " = " .. (v.adjAmount or "nil"))
			local currentVal = ret[v.statToAdjust] or 0 
			ret[v.statToAdjust] = currentVal + v.adjAmount
			applied = true
		end		
	end
	
	return ret, applied
end

function ItemScore:CalculateItemScore(itemTable, itemAdj, diff1Table, diff1Adj, 
									  diff2Table, diff2Adj, scoreSet, 
									  currentStats)
	-- apply adjustments if they apply to this scoreSet
	local trinketApplied = false
	itemTable, trinketApplied = self:ApplyAdjustments(itemTable, itemAdj, scoreSet)
	diff1Table, _ = self:ApplyAdjustments(diff1Table, diff1Adj, scoreSet)
	diff2Table, _= self:ApplyAdjustments(diff2Table, diff2Adj, scoreSet)
	
	local score = 0
	local score2 = 0
	
	local equippedScore = 0
	local equippedScore2 = 0
	
	local scoreSets = self:GetScoreSets()
	for i in pairs(self.supportedStats) do
		local thisScore = 0
		local thisCap = -1
		
		if scoreSet.info[i] then
			thisScore = scoreSet.info[i].score
			thisCap = scoreSet.info[i].cap
		elseif scoreSet.parent and scoreSets[scoreSet.parent] and 
			   scoreSets[scoreSet.parent].info[i] then
			thisScore = scoreSets[scoreSet.parent].info[i].score
			thisCap = scoreSets[scoreSet.parent].info[i].cap
		end 
		
		if thisScore ~= 0 then
			if not itemTable[i] then
				itemTable[i] = 0
			end
			-- calculate current stats for those with caps
			
			-- calculate for current equipment too
			local current1Val = 0
			local current2Val = 0
			
			if diff1Table then
				if not diff1Table[i] then
					-- no difference
					diff1Table[i] = 0
				end
				
				current1Val = diff1Table[i]
			else
				current1Val = 0	
			end
			
			current2Val = current1Val
			
			if diff2Table then
				if not diff2Table[i] then
					-- no difference
					diff2Table[i] = 0
				end
				
				current2Val = diff2Table[i]
			end
			
			if thisCap ~= -1 then
				-- calculate capped value
				-- get current stat (as needed)
				self:GetTotalCurrentStats(currentStats, i)
				
				-- determine current value(s) of currently equipped items
				local add = 0
				local equipAdd = 0
				
				-- get capped value
				add = getCappedValue(thisCap, currentStats[i] - current1Val, itemTable[i]) * thisScore
				score = score + add
				
				-- do for equipped item as well
				equipAdd = getCappedValue(thisCap, currentStats[i] - current1Val, current1Val) * thisScore
				equippedScore = equippedScore + equipAdd
			
				if current2Val ~= current1Val then
					add = getCappedValue(thisCap, currentStats[i] - current2Val, itemTable[i]) * thisScore
					
					-- do for equipped item as well
					equipAdd = getCappedValue(thisCap, currentStats[i] - current2Val, current2Val) * thisScore
				end
				
				score2 = score2 + add
				equippedScore2 = equippedScore2 + equipAdd

			else
			    local add = thisScore * itemTable[i]
				score = score + add 
				score2 = score2 + add
				equippedScore = equippedScore + (thisScore * current1Val)
				equippedScore2 = equippedScore2 + (thisScore * current2Val)
			end
		end   
	end
	
	return score, score2, equippedScore, equippedScore2, trinketApplied
end

function ItemScore:ShowItemCalculation(itemTable, scoreSet)
	local ret = ""
	
	local scoreSets = self:GetScoreSets()
	for i in pairs(self.supportedStats) do
		local thisScore = 0
		
		if scoreSet.info[i] then
			thisScore = scoreSet.info[i].score
		elseif scoreSet.parent and scoreSets[scoreSet.parent] and scoreSets[scoreSet.parent].info[i] then
			thisScore = scoreSets[scoreSet.parent].info[i].score
		end 
		
		if thisScore ~= 0 then
			if not itemTable[i] then
				itemTable[i] = 0
			end
			
			if itemTable[i] ~= 0 then
				if ret ~= "" then
					ret = ret .. " + "
				end
				ret = ret .. "(" .. itemTable[i] .. " * " .. thisScore .. ")"
			end	
		end   
	end
	
	return ret
end

function ItemScore:MakeTooltipText(scoreSet, score, score2, equippedScore, equippedScore2, trinketApplied)
	
	local trinketAppliedStr = ""
	if trinketApplied then
		trinketAppliedStr = "*"
	end
	
	local scoreText = string.format("%." .. scoreSet.precision .. "f", score)
			
	if score ~= score2 then
		scoreText = scoreText .. "/" .. string.format("%." .. scoreSet.precision .. "f", score2)
	end
	
	local diff = score - equippedScore
	local pctDiffStr = "inf%%"
	
	if score == 0 and equippedScore == 0 then
		pctDiffStr = "0%%"
	elseif equippedScore ~= 0 then
		local pctDiff = diff / equippedScore * 100.0
		pctDiffStr = string.format("%.0f", pctDiff) .. "%%"
	end
	
	local sign = ""
	local signColor = "|cFFFF0000"
	
	if diff >= 0 then
		sign = "+"
		signColor = "|cFF00FF00"
	end
	
	local equippedDiffText = signColor .. sign .. string.format("%." .. scoreSet.precision .. "f", diff) .. "|r"
	local equippedDiffPctText = signColor .. sign .. pctDiffStr .. "|r"
	
	if score ~= score2 or equippedScore ~= equippedScore2 then
		diff = score2 - equippedScore2
		if diff >= 0 then
			sign = "+"
			signColor = "|cFF00FF00"
		else
			sign = ""
			signColor = "|cFFFF0000"
		end
		
		pctDiffStr = "inf%%"
		if equippedScore2 ~= 0 then
			local pctDiff = diff / equippedScore2 * 100.0
			pctDiffStr = string.format("%.0f", pctDiff) .. "%%"
		end
			
		equippedDiffText = equippedDiffText .. "/" .. signColor ..  sign .. string.format("%." .. scoreSet.precision .. "f", diff) .. "|r"
		equippedDiffPctText = equippedDiffPctText .. "/" .. signColor ..  sign .. pctDiffStr .. "|r"
	end
			
	local ret = scoreSet.pattern
	
	ret = string.gsub(ret, "%%n", scoreSet.name)
	ret = string.gsub(ret, "%%a", scoreText)
	ret = string.gsub(ret, "%%d", equippedDiffText)
	ret = string.gsub(ret, "%%p", equippedDiffPctText)
	ret = string.gsub(ret, "%%t", trinketAppliedStr)
	
	return ret
end

-- adds table2 to table1
local function addTables(table1, table2)
	if not table2 or not table1 then return end
	table1 = table1 + table2
end

function ItemScore.ProcessTooltip(tooltip, name, link, ...)
	local currentStats = {}
	local itemTable = {}	
	
	-- get the item stats
	StatLogic:GetSum(tooltip, itemTable)
	
	-- check if we're a gem
	local isGem = nil
	if (not itemTable.itemType or itemTable.itemType == "") and itemTable.link then
		local _, _, _, _, _, itemType = GetItemInfo(itemTable.link)
		
		if itemType == "Gem" then
			isGem = true
		else
			return
		end
	end
	
	-- if we're not equippable or we are tabard, bag or quiver, then don't show scores
	if not isGem and itemTable and (itemTable.itemType == "" or itemTable.itemType == "INVTYPE_TABARD" or itemTable.itemType == "INVTYPE_BAG" or
		itemTable.itemType == "INVTYPE_QUIVER") then
		return
	end
	
	if itemTable.itemType and itemTable.itemType ~= "" then
		-- by default, we'll score and compare items in their unenchanted state
		local unchantedLink = StatLogic:RemoveEnchant(itemTable.link)
		
		if unchantedLink ~= itemTable.link then
			itemTable.link = unchantedLink
			StatLogic:GetSum(itemTable.link, itemTable)
		end
	end
	
	-- get difference in stat between equipped and this equipment
	--local equipDiff1, equipDiff2 = StatLogic:GetDiff(tooltip, nil, nil, true)
	local _, _, diff1Equip, diff2Equip = StatLogic:GetDiffID(tooltip, true)
	local diff1Table, diff2Table
	local diff1Adj, diff2Adj
	if diff1Equip then
		diff1Table = StatLogic:GetSum(diff1Equip, diff1Table)
		if itemTable.itemType == "INVTYPE_TRINKET" then
			diff1Adj = ItemScore:HandleTrinket(diff1Equip, diff1Table)
		end
	end
	if diff2Equip then
		diff2Table = StatLogic:GetSum(diff2Equip, diff2Table)
		if itemTable.itemType == "INVTYPE_TRINKET" then
			diff2Adj = ItemScore:HandleTrinket(diff2Equip, diff2Table)
		end
	end
	
	if itemTable.itemType == "INVTYPE_2HWEAPON" and diff2Table then
		-- sum the two tables
		addTables(diff1Table, diff2Table)
		diff2Table = nil
	end
	
	-- handle trinket scanning
	local itemAdj = nil
	if itemTable.itemType == "INVTYPE_TRINKET" then
		itemAdj = ItemScore:HandleTrinket(tooltip, itemTable)
	end
	
	local scoreSets = ItemScore:GetScoreSets()
	for i in pairs(scoreSets) do
		local scoreSet = scoreSets[i]
		if scoreSet.display then
			local setName = scoreSet.name
			
			-- calculate the score
			local score, score2, equippedScore, equippedScore2, applied = 
			ItemScore:CalculateItemScore(itemTable, itemAdj, diff1Table, 
										 diff1Adj, diff2Table, diff2Adj, 
										 scoreSet, currentStats)
			-- display the score
			local tooltipText = ItemScore:MakeTooltipText(scoreSet, score, 
														  score2, equippedScore, 
														  equippedScore2, 
														  applied)
			
			-- only do the debug string processing if it exists in the string
			if string.find(tooltipText, "%%x") then
				local debugTxt = ItemScore:ShowItemCalculation(itemTable, 
															   scoreSet)
				tooltipText = string.gsub(tooltipText, "%%x", debugTxt)
			end
			
			tooltip:AddLine(tooltipText, scoreSet.r, scoreSet.g, scoreSet.b)
			tooltip:Show()
		end
	end
end

function ItemScore:PrintTotalScores()
	local currentStats = {}
	self:Print("Total scores:")
	
	local scoreSets = self:GetScoreSets()
	for j in pairs(scoreSets) do
		local totalScore = 0
		local scoreSet = scoreSets[j]
		
		for i in pairs(self.supportedStats) do
			local thisScore = 0
			local thisCap = -1
			
			if scoreSet.info[i] then
				thisScore = scoreSet.info[i].score
				thisCap = scoreSet.info[i].cap
			elseif scoreSet.parent then
				local parentSet = scoreSets[scoreSet.parent]
				if parentSet and parentSet.info[i] then
					thisScore = parentSet.info[i].score
					thisCap = parentSet.info[i].cap
				end
			end 
			if thisScore ~= 0 then
				self:GetTotalCurrentStats(currentStats, i)
				
				if thisCap ~= -1 then
					totalScore = totalScore + (thisScore * getCappedValue(thisCap, 0, currentStats[i]))
				else
					totalScore = totalScore + (thisScore * currentStats[i])
				end
			end
		end
		
		self:Print(scoreSet.name .. ": " .. string.format("%." .. scoreSet.precision .. "f", totalScore))
	end
end

function ItemScore:OnCmd(args)
	local cmd = string.split(' ', args):lower() or args:lower()
	local restOfString = string.gsub(args, cmd .. ' ', '')

	self:PrintTotalScores()
end
